/*
 * Copyright (c) 2008 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyneConfig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_POCOLIBS
#include <portLib.h>
#endif
#include <velodyne/velodyneLib.h>
#include "velodyne_private.h"

/* Calibration data file, built from the XML data file provided 
   with each sensor */
#include "velodyneCalibration.h"

/*
 * Precompute the calibration parameters to avoid re-doing it for each point
 * when converting them in Cartesian coordinates
 */

#define VLS_NUM_ROT_ANGLES  36000
#define VLS_NUM_LASERS  64

double cosHeadingAngleTable[VLS_NUM_ROT_ANGLES][VLS_NUM_LASERS];
double sinHeadingAngleTable[VLS_NUM_ROT_ANGLES][VLS_NUM_LASERS];

const double VELODYNE_OFFSET_SCALE = 0.01;
const double VELODYNE_HEADING_SCALE = 0.01;

double cosVertAngleTable[VLS_NUM_LASERS];
double sinVertAngleTable[VLS_NUM_LASERS];
double vertOffsetTable[VLS_NUM_LASERS];
double horizOffsetTable[VLS_NUM_LASERS];

void
velodyne_init_calibration(void)
{
  int		i, j;
  double	headingAngle;

  for (i = 0; i < VLS_NUM_ROT_ANGLES; i++) {
    for (j = 0; j < VLS_NUM_LASERS; j++) {
      headingAngle = 
	((i * VELODYNE_HEADING_SCALE) - vc[j].rotCorrection) * M_PI / 180.0;
      cosHeadingAngleTable[i][j] = cos(headingAngle);
      sinHeadingAngleTable[i][j] = sin(headingAngle);
    }
  }
  
  for (j = 0; j < VLS_NUM_LASERS; j++) {
    cosVertAngleTable[j] 
      = cos(vc[j].vertCorrection * M_PI / 180.0);
    sinVertAngleTable[j] 
      = sin(vc[j].vertCorrection * M_PI / 180.0);
    vertOffsetTable[j] 
      = vc[j].vertOffsetCorrection * VELODYNE_OFFSET_SCALE;
    horizOffsetTable[j] 
      = vc[j].horizOffsetCorrection * VELODYNE_OFFSET_SCALE;
  }
}


/*
 * Function to convert raw velodyne data into cartesian coordinates in the
 * sensor frame. See the Velodyne documentation or reference [1] for explanation
 * of the way the conversion is achieved.
 *
 * [1]: N. Muhammad and S. Lacroix. "Calibration of a rotating multi-beam
 * Lidar", IEEE/RSJ International Conference on Intelligent Robots and Systems,
 * 2010. 
 */

const double VELODYNE_DIST_SCALE = 0.002;

static inline void
velodyne_compute_coords(u_int16_t laserNum,
			u_int16_t heading, 
			u_int16_t dist, 
			u_int8_t intensity,
			struct VELODYNE_POINT *point)
{
  struct VELODYNE_CALIBRATION cal = vc[laserNum];

  double distance =  dist * VELODYNE_DIST_SCALE 
    + cal.distCorrection * VELODYNE_OFFSET_SCALE;

  if ((dist == 0) 
      || (distance < VELODYNE_MIN_DISTANCE) 
      || (distance > VELODYNE_MAX_DISTANCE)) {
    point->status = VELODYNE_BAD_POINT;
    return;
  }

  double cosVertAngle = cosVertAngleTable[laserNum];
  double sinVertAngle = sinVertAngleTable[laserNum];
  double vertOffset = vertOffsetTable[laserNum];
  double horizOffset = horizOffsetTable[laserNum];
  double cosHeadingAngle = cosHeadingAngleTable[heading][laserNum];
  double sinHeadingAngle = sinHeadingAngleTable[heading][laserNum];

  double xyDistance = distance * cosVertAngle - vertOffset * sinVertAngle;

  point->status = VELODYNE_GOOD_POINT;
  /* Warning: The two lines above take into account the fact that the original
     sensor frame is as follows: the Y axis is oriented opposite to the
     wire connector, the heading is 0 along this Y axis, and the heading is
     positive clock-wise (strange convention fron Velodyne). Doing x = y, and
     y = -x defines a usual sensor frame, with the X axis heading opposite to
     the wire connector */
  point->x = xyDistance * cosHeadingAngle + horizOffset * sinHeadingAngle;
  point->y = - (xyDistance * sinHeadingAngle - horizOffset * cosHeadingAngle);
  point->z = distance * sinVertAngle + vertOffset * cosVertAngle;
  point->intensity = intensity;
}


/*----------------------------------------------------------------------*/

/**
 * \brief Get access to the lidar device 
 *
 * name is either "udp" to read live packets from the sensor, or
 * the name of a file containing data captured using tcpdump
 */
VELODYNE_INTERFACE *
velodyne_open(char *name)
{
	VELODYNE_INTERFACE *dev;

	dev = (VELODYNE_INTERFACE *)malloc(sizeof(VELODYNE_INTERFACE));
	if (dev == NULL) {
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_NO_MEMORY);
#endif
		return NULL;
	}
	if (strcmp(name, "udp") == 0) {
		dev->fd = velodyne_setup_socket();
		if (dev->fd == -1) {
			free(dev);
			return NULL;
		}
		dev->type = VELODYNE_NETWORK;
	} else {
		if (velodyne_pcap_file_open(name, &dev->pcap) == -1) {
			free(dev);
			return NULL;
		}
		dev->type = VELODYNE_FILE;
	}
	velodyne_init_calibration();
	return dev;
}

/*----------------------------------------------------------------------*/

/**
 * \brief Terminate the access to the device
 */
void
velodyne_close(VELODYNE_INTERFACE *device)
{
	switch (device->type) {
	case VELODYNE_FILE:
		velodyne_close_pcap(device->pcap);
	case VELODYNE_NETWORK:
		velodyne_close_socket(device->fd);
#ifdef HAVE_POCOLIBS
	default:
		errnoSet(S_velodyneLib_INVALID_DEVICE);
#endif
	}
	free(device);
}

/*----------------------------------------------------------------------*/
/**
 * \brief Read a packet from the device 
 */
int
velodyne_read_packet(VELODYNE_INTERFACE *device, VELODYNE_PACKET_STR *p)
{
	switch (device->type) {
	case VELODYNE_FILE:
		return velodyne_pcap_read_packet(device->pcap, p);
	case VELODYNE_NETWORK:
		return velodyne_udp_rcv(device->fd, p, 0);
	default:
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_INVALID_DEVICE);
#endif
		return -1;
	}
	return 0;
}

/*
 * A packet of data contains VELODYNE_NUM_SHOTS blocks of VELODYNE_NUM_LASERS
 * points (i.e. 12 x 32 = 384 points). Each block comes either for the upper or
 * the lower sets of lasers (the sensor has in total 2 * VELODYNE_NUM_SHOTS =
 * VLS_NUM_LASERS = 64 lasers), and is encoded in the structure
 * VELODYNE_FIRE_STR, that indicates:
 * - whether it is an upper or lower block
 * - its rotational position, i.e. the heading angle of the rotating sensor when
 * data are acquired
 * - VELODYNE_NUM_SHOTS structures VELODYNE_LASER_STR
 * 
 */

void
velodyne_print_packet(VELODYNE_PACKET_STR *p)
{
	int i, j;
	VELODYNE_RAW_PACKET_STR *packet = &p->data;
	struct tm localtime;
	char buf[100];

	localtime_r(&p->ts.tv_sec, &localtime);
	strftime(buf, sizeof(buf), "%F %T", &localtime);
	printf("%s.%06u\n", buf, (unsigned long)p->ts.tv_usec);
	for (i = 0; i < VELODYNE_NUM_SHOTS; i++) {
		printf("Block %d %s, %d:\n", i, 
		    packet->shots[i].lower_upper == VELODYNE_UPPER_HEADER ?
		    "upper" : 
		    packet->shots[i].lower_upper == VELODYNE_LOWER_HEADER ? 
		    "lower" : "unknown",
		    packet->shots[i].rotational_pos);
		printf("Distance: ");
		for (j = 0; j < VELODYNE_NUM_LASERS; j++)
			printf("%d\t", packet->shots[i].lasers[j].distance);
		printf("\nIntensity: ");
		for (j = 0; j < VELODYNE_NUM_LASERS; j++)
			printf("%d\t", packet->shots[i].lasers[j].intensity);
		printf("\n");
	}
	printf("Status: %f %s", packet->degrees + packet->frac_degrees/256.0,
	    packet->status_string);
	printf("\n\n");
}

/*----------------------------------------------------------------------*/

void
velodyne_print_packet_3d(VELODYNE_PACKET_STR *pkt)
{
  VELODYNE_RAW_PACKET_STR *packet = &pkt->data;
  struct VELODYNE_POINT p;
  struct VELODYNE_FIRE_STR *s;
  int i, j, n;
  
  for (i = 0; i < VELODYNE_NUM_SHOTS; i++) {
    s = &(packet->shots[i]);
    for (j = 0; j < VELODYNE_NUM_LASERS; j++) {
      n = (s->lower_upper == VELODYNE_UPPER_HEADER ? 0 : 1)
	* VELODYNE_NUM_LASERS + j;
      velodyne_compute_coords(n, s->rotational_pos, 
			      s->lasers[j].distance, s->lasers[j].intensity,
			      &p);
	printf("%lf %lf %lf %d\n",
	       p.x, p.y, p.z, s->lasers[j].intensity); 
   }
  }
}

/*
 * Converts a VELODYNE_PACKET_STR into a VELODYNE_CARTESIAN_PACKET_STR. Points
 * are expressed in the sensor frame.
 */

void
velodyne_convert_packet(VELODYNE_PACKET_STR *pkt,
			VELODYNE_CARTESIAN_PACKET_STR *cartesianPkt)
{
  VELODYNE_RAW_PACKET_STR *packet = &pkt->data;
  struct VELODYNE_POINT* p;
  struct VELODYNE_FIRE_STR *s;
  int i, j, n;
  
  p = cartesianPkt->points;
  for (i = 0; i < VELODYNE_NUM_SHOTS; i++) {
    s = &(packet->shots[i]);
    for (j = 0; j < VELODYNE_NUM_LASERS; j++) {
      n = (s->lower_upper == VELODYNE_UPPER_HEADER ? 0 : 1)
	* VELODYNE_NUM_LASERS + j;
      velodyne_compute_coords(n, s->rotational_pos, 
			      s->lasers[j].distance, s->lasers[j].intensity,
			      p);
      p++;
    }
  }
  cartesianPkt->ts = pkt->ts;
}

